#!/bin/bash

if [ $# -lt 5 ]
then
	echo 
        echo "Usage: mongo.sh <filesystem> <device> <mount_point> <log> <repeat>"
	echo
	echo "<filesystem> - the name of filesystem [reiserfs|ext2]"
	echo "<device>     - the device for benchmark (e.g. /dev/hda9)"
	echo "<mount_point>   - mount-point for the filesystem (e.g. /mnt/testfs)"
	echo "<log>        - the name prefix for file of benchmark results"
	echo "<repeat>     - the number of benchmark repetitions"
 	echo
	echo examples:
	echo mongo.sh reiserfs /dev/hda9 /mnt/testfs journaling_on_results 1
	echo mongo.sh ext2 /dev/hda9 /mnt/testfs ext2_results 1
	echo
	echo ...............................................
	echo The results will be put in ./results directory
	echo
	exit
fi

#...You can skip running part of the benchmark by setting "0" instead of "1":
#----------------------------------------------------------------------------
execute_single_process_phase=1

    execute_copying_single=1
    execute_scanning_single=1
    execute_directory_stats_single=1
    execute_deleting_single=1
    
execute_multi_process_phase=1

    execute_copying_multi=1
    execute_scanning_multi=1
    execute_directory_stats_multi=1
    execute_deleting_multi_parallel=0
    execute_deleting_multi=1

#-------- reiser_fract_tree parameters----------------
# defines
x1mb=$[1024*1024]
x5mb=$[5 * $x1mb]
x50mb=$[50 * $x1mb]
x100mb=$[100 * $x1mb]

# Total amount of bytes in all files on test partition
small_bytes=$x5mb
medium_bytes=$x50mb
large_bytes=$x100mb

# Median size of files in bytes for first tree to create
small_size=100
medium_size=1000
large_size=100000

#Keep the largest file to one fifth (100 million bytes) of the total tree size.
max_file_size=100000000
#Yuri Shevchuk says that 0 is the median size in real life, so I believe him.
median_dir_nr_files=0
#This should be larger, change once done testing.

bytes_to_consume=10000000

median_file_size=100 
max_file_size=1000000

median_dir_nr_files=100
max_directory_nr_files=10000 

median_dir_branching=0
max_dir_branching=1 

#This should be varying, someday....
write_buffer_size=4096

numb_of_bytes=($small_bytes $medium_bytes $large_bytes)
size_of_files=($small_size $medium_size $large_size)
#level_of_tree=(2 2 2)
#fract_of_tree=(5 5 5)
reiser_fract_tree_rep_counter=3
total_params=${#numb_of_bytes[*]}


#... Set working directories
#-----------------------------
TOPDIR=$(pwd)
RESDIR=$TOPDIR/results
HTMLDIR=$RESDIR/html

FILESYSTEM=$1
DEVICE=$2
TESTDIR=$3
LOGFILE=$RESDIR/$4
LOGFILE2=${LOGFILE}_table
LOGFILE3=${LOGFILE}.tbl
REPETITIONS=$5


TMPFILE=$RESDIR/mongo_tmp

rep_left=$REPETITIONS

READIT=$TOPDIR/mongo_read


#... Make directories for results
#--------------------------------
if ! [ -d $RESDIR ]
then
    echo Creating dir: $RESDIR
    mkdir $RESDIR
fi
if ! [ -d $HTMLDIR ]
then
    echo Creating dir: $HTMLDIR
    mkdir $HTMLDIR
fi

#... Compile *.c files if it is necessary
#----------------------------------------
if ! [ -f reiser_fract_tree ]
then
    echo Compiling reiser_fract_tree.c ...
    gcc reiser_fract_tree.c -o reiser_fract_tree -lm
fi

if ! [ -f mongo_read ]
then
    echo Compiling mongo_read.c ...
    gcc mongo_read.c -o mongo_read
fi

if ! [ -f mongo_compare ]
then
    echo Compiling mongo_compare.c ...
    gcc mongo_compare.c -o mongo_compare
fi

#... Check the command string parameters
#---------------------------------------
if [ $FILESYSTEM = "reiserfs" ] || [ $FILESYSTEM = "ext2" ]
then
    echo
else
    echo "mongo.sh: not valid filesystem name: $FILESYSTEM"
    echo "Usage: mongo.sh <filesystem> <device> <mount_point> <log> <repeat>"
    exit 1
fi

if ! [ -b $DEVICE ]
then
    echo "mongo.sh: not valid device: $DEVICE"
    echo "Usage: mongo.sh <filesystem> <device> <mount_point> <log> <repeat>"
    exit 1
fi


#... ------------------The functions definitions --------------------

function copy() {
    cp -r $1 $2
}

function remove() {
    rm -rf $@
}

function make_fsys() {

    echo "FILESYSTEM=$FILESYSTEM" >> $LOGFILE

    umount $TESTDIR

    if [ $FILESYSTEM = "reiserfs" ]
    then
	echo y | mkreiserfs $DEVICE; echo
	mount -t reiserfs $DEVICE $TESTDIR
    fi

    if [ $FILESYSTEM = "ext2" ]
    then
	echo y | mke2fs $DEVICE
	mount -t ext2 $DEVICE $TESTDIR
    fi
}

function cmd_create() {
        $TOPDIR/reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size $TESTDIR
}

function mongo_single_process() {

    # make and mount the file system
    make_fsys;

    echo
    echo "mongo_single_process, the_set_of_param.N=$par_set_n of $total_params"
    echo "Results in file : $LOGFILE"
    echo

    echo 0.Creating files of median size $median_file_size bytes ...
    #echo "********* Phase 0: Creating $files_total files of base size $median_file_size in $dirs_total dirs *********" >> $LOGFILE
    echo "********* Phase 0: Creating files of median size $median_file_size bytes *********" >> $LOGFILE
    i=0; total=0
    while [ $i -lt $reiser_fract_tree_rep_counter ]
    do
	#echo "reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size"
	echo -n "Creating : "
        echo -n "Creating : " >> $LOGFILE	
        t=`(time -p $TOPDIR/reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $median_dir_nr_files $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size $TESTDIR/testdir-$i 0) 2>&1`
	creating=`(echo $t | (cut -d" " -f2) 2>&1)`
	echo "$creating sec." >> $LOGFILE
	echo "$creating	sec."
	total=`export total; export creating; perl -e 'print $ENV{'total'} + $ENV{'creating'}'`	
        i=$[ $i + 1 ]
    done
    echo "total creating time: $total sec."
    echo "total creating time: $total sec." >> $LOGFILE
    creating=$total
            
    u=`du -sk $TESTDIR 2>&1`
    used=`(echo $u | (cut -d" " -f1) 2>&1)`
    echo "Used disk space : $used KB"
    echo "Used disk space : $used KB" >> $LOGFILE
    
    sync

    if [  $execute_copying_single -eq 1 ]
    then
	echo 1.Copying files...
	echo "********* Phase I: Copying files *********" >> $LOGFILE
	i=0;total=0;
	while [ $i -lt $reiser_fract_tree_rep_counter ]
	do
	    echo -n "Copying  : "
    	    echo -n "Copying  : " >> $LOGFILE
	    t=`(time -p copy $TESTDIR/testdir-$i $TESTDIR/testdir-$i-$i ) 2>&1`
	    copying=`(echo $t | (cut -d' ' -f2) 2>&1)`
	    echo "$copying sec.">> $LOGFILE
            echo "$copying sec."
	    total=`export total; export copying; perl -e 'print $ENV{'total'} + $ENV{'copying'}'`	
	    i=$[ $i + 1 ]
	done
        echo "total copying time: $total sec."
	echo "total copying time: $total sec." >> $LOGFILE
	copying=$total

        files_total=`find $TESTDIR -type f | wc -l`
        dirs_total=`find $TESTDIR -type d | wc -l`
	
        u=`du -sk $TESTDIR 2>&1`
	used2=`(echo $u | (cut -d" " -f1) 2>&1)`

	echo "Total files : $files_total"
	echo "Total dirs  : $dirs_total"

	echo "Used disk space (du): $used2 KB"
	echo "Used disk space (du): $used2 KB" >> $LOGFILE
		
	sync
    fi


    df >> $LOGFILE

    if [ $execute_scanning_single -eq 1 ]
    then
	echo 2.Scanning each file...
	echo "********* Phase II: Scanning each file *********" >> $LOGFILE
	echo -n "Scanning : "
    	echo -n "Scanning : " >> $LOGFILE
	#t=`(time -p find $TESTDIR -type f -exec $READIT {} $write_buffer_size >& /dev/null \; ) 2>&1`
        t=`(time -p find $TESTDIR -type f |xargs $READIT ) 2>&1`	
	scanning=`(echo $t | (cut -d' ' -f2) 2>&1)`
	echo "$scanning sec." >> $LOGFILE
	echo "$scanning sec."
	sync
    fi
    
     
    if [ $execute_directory_stats_single -eq 1 ]
    then
	echo 3.Recursive directory stats...
	echo "********* Phase III: Recursive directory stats *********" >> $LOGFILE
	echo -n "Stats time  : "
    	echo -n "Stats time  : " >> $LOGFILE
	#t=`(time -p find $TESTDIR  -exec ls -l {} > /dev/null \;  ) 2>&1`
	t=`(time -p find $TESTDIR  | xargs ls -l > /dev/null ) 2>&1`
	(time -p du -s $TESTDIR >> $LOGFILE ) >> $LOGFILE 2>&1
	stating=`(echo $t | (cut -d' ' -f2) 2>&1)`
	echo "$stating sec." >> $LOGFILE
	echo "$stating sec."
	sync
    fi

    if [ $execute_deleting_single -eq 1 ]
    then
	echo 4.Deleting remaining files...
	echo "********* Phase VIIII: Deleting remaining files ****" >> $LOGFILE
	echo -n "Deleting : "
    	echo -n "Deleting : " >> $LOGFILE
	t=`(time -p remove $TESTDIR/* ) 2>&1`
	deleting=`(echo $t | (cut -d' ' -f2) 2>&1)`
	echo "$deleting sec.">> $LOGFILE
	echo "$deleting sec."
	sync
    fi

    echo >> $LOGFILE2
    echo "MONGO_SINGLE_PROCESS BENCHMARK RESULTS (time in sec.)" >> $LOGFILE2
    echo "  FILESYSTEM=$FILESYSTEM" >> $LOGFILE2
    echo "  parameters: files=$files_total, base_size=$median_file_size bytes, dirs=$dirs_total" >> $LOGFILE2    
    echo -e "-----------------------------------------------------" >> $LOGFILE2
    echo -e "Create\tCopy\tRead\tStats\tDelete" >> $LOGFILE2
    echo -e " time \ttime\ttime\ttime \t time " >> $LOGFILE2
    echo -e "-----------------------------------------------------" >> $LOGFILE2
    echo -e "$creating\t$copying\t$scanning\t$stating\t$deleting" >> $LOGFILE2
    echo -e "-----------------------------------------------------" >> $LOGFILE2
    echo "The size of files tree : " >> $LOGFILE2
    echo "         after create = $used kb" >> $LOGFILE2
    echo "         after copy   = $used2 kb" >> $LOGFILE2 
    echo >> $LOGFILE2

    echo >> $LOGFILE3
    echo "MONGO_SINGLE_PROCESS      " >> $LOGFILE3  
    echo "parameters:               " >> $LOGFILE3
    echo "files=$files_total        " >> $LOGFILE3
    echo "base_size=$median_file_size bytes     " >> $LOGFILE3
    echo "dirs=$dirs_total          " >> $LOGFILE3
    echo >> $LOGFILE3 
    
    echo "FSYS=$FILESYSTEM          " >> $LOGFILE3
    echo "(time in sec.)            " >> $LOGFILE3
    echo "Create         : $creating" >> $LOGFILE3
    echo "Copy           : $copying " >> $LOGFILE3
    echo "Read           : $scanning" >> $LOGFILE3
    echo "Stats          : $stating " >> $LOGFILE3
    echo "Delete         : $deleting" >> $LOGFILE3    
    echo >> $LOGFILE3

        
    echo "******* The end of mongo_single_process *******" >> $LOGFILE

}

function del_tmpfiles() {

    i=0; 
    while [ $i -le $reiser_fract_tree_rep_counter ]
    do
	if [ -f $TMPFILE-$i ]
	then
	    rm $TMPFILE-$i
	fi    
	i=$[ $i + 1 ]
    done
    
    if [ -f $TMPFILE ]
    then
	rm $TMPFILE
    fi
    
}

function mongo_multi_process() {

    # make and mount the file system
    make_fsys;

    echo
    echo "mongo_multi_process, rep.N=$rep_n of $REPETITIONS, the_set_of_param.N=$par_set_n of $total_params"
    echo "Results in file : $LOGFILE"
    echo

    #echo 0..Creating $files_total files of base size $median_file_size bytes in $dirs_total dirs ...
    echo 0.Creating files of median size $median_file_size bytes ...
    #echo "********* Phase 0: Creating $files_total files of base size $median_file_size in $dirs_total dirs *********" >> $LOGFILE
    echo "********* Phase 0: Creating files of median size $median_file_size bytes *********" >> $LOGFILE
    i=1
    while [ $i -lt $reiser_fract_tree_rep_counter ]
    do
        #echo "reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size" >> $LOGFILE
	#echo "reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size"
	echo "reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $median_dir_nr_files $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size" >> $LOGFILE
	echo "reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $median_dir_nr_files $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size"
        #(time -p $TOPDIR/reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size 0) > $TMPFILE-$i 2>&1 &	
        (time -p $TOPDIR/reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $median_dir_nr_files $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size $TESTDIR/testdir-$i 0) > $TMPFILE-$i 2>&1 &
	i=$[ $i + 1 ]
    done
    i=$reiser_fract_tree_rep_counter
    creation_cmd="(time -p $TOPDIR/reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size 0) >> $LOGFILE 2>&1";
    #echo "reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size" >> $LOGFILE
    #echo "reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size"
    echo "reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $median_dir_nr_files $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size" >> $LOGFILE 
    echo "reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $median_dir_nr_files $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size"

    #(time -p $TOPDIR/reiser_fract_tree $files $level $median_file_size $inverse $TESTDIR/testdir-$i $write_buffer_size 0) > $TMPFILE-$i 2>&1
    (time -p $TOPDIR/reiser_fract_tree $bytes_to_consume $median_file_size $max_file_size $median_dir_nr_files $max_directory_nr_files $median_dir_branching $max_dir_branching $write_buffer_size $TESTDIR/testdir-$i 0) > $TMPFILE-$i 2>&1 
    sleep 30
    
    total=0
    i=1
    while [ $i -le $reiser_fract_tree_rep_counter ]
    do
	echo -n "Creating : "
        echo -n "Creating : " >> $LOGFILE
	creating=`(cat $TMPFILE-$i | grep "real" | (cut -d" " -f2) 2>&1)`
	echo "$creating sec." >> $LOGFILE
	echo "$creating	sec."
	total=`export total; export creating; perl -e 'print $ENV{'total'} + $ENV{'creating'}'`	
        i=$[ $i + 1 ]
    done
    echo "total creating time: $total sec."
    echo "total creating time: $total sec." >> $LOGFILE
    creating=$total

    del_tmpfiles;
    
    u=`du -sk $TESTDIR 2>&1`
    used=`(echo $u | (cut -d" " -f1) 2>&1)`
    echo "Used disk space : $used KB"
    echo "Used disk space : $used KB" >> $LOGFILE
    
    sync

    if [  $execute_copying_multi -eq 1 ]
    then
	echo 1..Copying files...
	echo "********* Phase I: Copying files *********" >> $LOGFILE
	i=1
	while [ $i -lt $reiser_fract_tree_rep_counter ]
	do
	    (time -p copy $TESTDIR/testdir-$i $TESTDIR/testdir-$i-$i ) >  $TMPFILE-$i 2>&1 &
	    i=$[ $i + 1 ]
	done
	i=$reiser_fract_tree_rep_counter
	(time -p copy $TESTDIR/testdir-$i $TESTDIR/testdir-$i-$i ) >  $TMPFILE-$i 2>&1

	sleep 30
	
	total=0
	i=1
	while [ $i -le $reiser_fract_tree_rep_counter ]
	do
	    echo -n "Copying : "
    	    echo -n "Copying : " >> $LOGFILE
	    copying=`(cat $TMPFILE-$i | grep "real" | (cut -d" " -f2) 2>&1)`
	    echo "$copying sec." >> $LOGFILE
	    echo "$copying	sec."
	    total=`export total; export copying; perl -e 'print $ENV{'total'} + $ENV{'copying'}'`	
    	    i=$[ $i + 1 ]
	done	
	echo "total copying time: $total sec."
	echo "total copying time: $total sec." >> $LOGFILE
	copying=$total
	del_tmpfiles;

        files_total=`find $TESTDIR -type f | wc -l`
        dirs_total=`find $TESTDIR -type d | wc -l`
	
        u=`du -sk $TESTDIR 2>&1`
	used2=`(echo $u | (cut -d" " -f1) 2>&1)`
	echo "Total files : $files_total"
	echo "Total dirs  : $dirs_total"
	echo "Used disk space (du) : $used2 KB"
	echo "Used disk space (du) : $used2 KB" >> $LOGFILE	
	sync
    fi

    df >> $LOGFILE

    if [ $execute_scanning_multi -eq 1 ]
    then
	echo 2.a..Scanning each file--- parallel...
	echo "********* Phase II-a: Scanning each file--- parallel *********" >> $LOGFILE
	if [  $execute_copying_multi -eq 1 ]
	then
	    i=1
	    while [ $i -lt $reiser_fract_tree_rep_counter ]
	    do
		#(time -p find $TESTDIR/testdir-$i $TESTDIR/testdir-$i-$i -type f -exec $READIT {} $write_buffer_size >& /dev/null \; ) >  $TMPFILE-$i 2>&1 &
		(time -p find $TESTDIR/testdir-$i $TESTDIR/testdir-$i-$i -type f | xargs $READIT ) >  $TMPFILE-$i 2>&1 &
		i=$[ $i + 1 ]
	    done
	    i=$reiser_fract_tree_rep_counter
	    #(time -p find $TESTDIR/testdir-$i $TESTDIR/testdir-$i-$i -type f -exec $READIT {} $write_buffer_size >& /dev/null \; ) >  $TMPFILE-$i 2>&1
	    (time -p find $TESTDIR/testdir-$i $TESTDIR/testdir-$i-$i -type f |xargs $READIT ) >  $TMPFILE-$i 2>&1
	else
	    i=1
	    while [ $i -lt $reiser_fract_tree_rep_counter ]
	    do
    		#(time -p find $TESTDIR/testdir-$i -type f -exec $READIT {} $write_buffer_size >& /dev/null \; ) >  $TMPFILE-$i 2>&1 &
		(time -p find $TESTDIR/testdir-$i -type f | xargs $READIT ) >  $TMPFILE-$i 2>&1 &
		i=$[ $i + 1 ]
	    done
	    i=$reiser_fract_tree_rep_counter
	    #(time -p find $TESTDIR/testdir-$i -type f -exec $READIT {} $write_buffer_size >& /dev/null \; ) >  $TMPFILE-$i 2>&1
	    (time -p find $TESTDIR/testdir-$i -type f | xargs $READIT ) >  $TMPFILE-$i 2>&1
	fi
	sleep 30
	
	total=0
	i=1
	while [ $i -le $reiser_fract_tree_rep_counter ]
	do
	    echo -n "Scanning_a : "
    	    echo -n "Scanning_a : " >> $LOGFILE
	    scanning=`(cat $TMPFILE-$i | grep "real" | (cut -d" " -f2) 2>&1)`
	    echo "$scanning sec." >> $LOGFILE
	    echo "$scanning sec."
	    total=`export total; export scanning; perl -e 'print $ENV{'total'} + $ENV{'scanning'}'`	
    	    i=$[ $i + 1 ]
	done	
	echo "total scanning_a time: $total sec."
	echo "total scanning_a time: $total sec." >> $LOGFILE
	scanning=$total    
	del_tmpfiles;
	sync

	echo 2.b..Scanning each file --- serial after parallel...
	echo "********* Phase II-b: Scanning each file --- serial after parallel *********" >> $LOGFILE
	#(time -p find $TESTDIR -type f -exec $READIT {} $write_buffer_size >& /dev/null \; ) > $TMPFILE 2>&1
        (time -p find $TESTDIR -type f |xargs $READIT {} ) > $TMPFILE 2>&1
	sleep 30
	
	echo -n "Scanning_b : "
    	echo -n "Scanning_b : " >> $LOGFILE
	scanning_b=`(cat $TMPFILE | grep "real" | (cut -d" " -f2) 2>&1)`
	echo "$scanning_b sec." >> $LOGFILE
	echo "$scanning_b sec."
	del_tmpfiles;
	sync
    fi

    if [ $execute_directory_stats_multi -eq 1 ]
    then
	echo 3..Recursive directory stats...
	echo "********* Phase III: Recursive directory stats *********" >> $LOGFILE
	#(time -p find $TESTDIR -exec ls -l {} > /dev/null \;  ) >  $TMPFILE 2>&1
	(time -p find $TESTDIR |xargs ls -l > /dev/null ) >  $TMPFILE 2>&1
	#(time -p du -s $TESTDIR >> $LOGFILE ) >> $LOGFILE 2>&1
	sleep 30
	echo -n "Stats time : "
    	echo -n "Stats time : " >> $LOGFILE
	stating=`(cat $TMPFILE | grep "real" | (cut -d" " -f2) 2>&1)`
	echo "$stating sec." >> $LOGFILE
	echo "$stating sec."
	del_tmpfiles;
	sync
    fi
    
    if [ $execute_deleting_multi_parallel -eq 1 ]
    then
	echo 4..Deleting remaining files parallel...
	echo "********* Phase VIIII: Deleting remaining files ****" >> $LOGFILE
	    i=1
	    while [ $i -lt $reiser_fract_tree_rep_counter ]
	    do
		(time -p remove $TESTDIR/testdir-$i ) > $TMPFILE-$i 2>&1 &
		i=$[ $i + 1 ]
	    done
	    i=$reiser_fract_tree_rep_counter
	    (time -p remove $TESTDIR/testdir-$i ) > $TMPFILE-$i 2>&1 &	    
	    
	sleep 30
	
	total=0
	i=1
	while [ $i -le $reiser_fract_tree_rep_counter ]
	do
	    echo -n "Deleting   : "
    	    echo -n "Deleting   : " >> $LOGFILE
	    deleting=`(cat $TMPFILE-$i | grep "real" | (cut -d" " -f2) 2>&1)`
	    echo "$deleting sec." >> $LOGFILE
	    echo "$deleting sec."
	    total=`export total; export deleting; perl -e 'print $ENV{'total'} + $ENV{'deleting'}'`	
    	    i=$[ $i + 1 ]
	done	
	echo "total deleting time: $total sec."
	echo "total deleting time: $total sec." >> $LOGFILE
	deleting=$total    
	del_tmpfiles;
	sync

    fi
    
    if [ $execute_deleting_multi -eq 1 ]
    then
	echo 4..Deleting remaining files...
	echo "********* Phase VIIII: Deleting remaining files ****" >> $LOGFILE
	(time -p remove $TESTDIR/* ) > $TMPFILE 2>&1
	echo -n "Deleting : "
    	echo -n "Deleting : " >> $LOGFILE
	deleting=`(cat $TMPFILE | grep "real" | (cut -d" " -f2) 2>&1)`
	echo "$deleting sec." >> $LOGFILE
	echo "$deleting sec."
	del_tmpfiles;
	sync

    fi
    
    echo >> $LOGFILE2
    echo "MONGO_MULTI_PROCESS BENCHMARK RESULTS (time in sec.)" >> $LOGFILE2
    echo "  FILESYSTEM=$FILESYSTEM" >> $LOGFILE2
    echo "  parameters: files=$files_total, base_size=$median_file_size bytes, dirs=$dirs_total" >> $LOGFILE2
    echo -e "----------------------------------------------------" >> $LOGFILE2
    echo -e "Create\tCopy\tRead  \tRead  \tStats\tDelete" >> $LOGFILE2
    echo -e " time \ttime\tparall\tserial\ttime \t time " >> $LOGFILE2    
    echo -e "----------------------------------------------------" >> $LOGFILE2
    echo -e "$creating\t$copying\t$scanning\t$scanning_b\t$stating\t$deleting" >> $LOGFILE2
    echo -e "----------------------------------------------------" >> $LOGFILE2
    echo "The size of files tree : " >> $LOGFILE2
    echo "         after create = $used kb" >> $LOGFILE2
    echo "         after copy   = $used2 kb" >> $LOGFILE2 
    echo >> $LOGFILE2


    echo >> $LOGFILE3

    echo "MONGO_MULTI_PROCESS      " >> $LOGFILE3 
    echo "parameters:              " >> $LOGFILE3
    echo "files=$files_total       " >> $LOGFILE3
    echo "base_size=$median_file_size bytes    " >> $LOGFILE3
    echo "dirs=$dirs_total         " >> $LOGFILE3
    echo >> $LOGFILE3 
    
    echo "FSYS=$FILESYSTEM         " >> $LOGFILE3
    echo "(time in sec.)           " >> $LOGFILE3
    echo "Create        : $creating" >> $LOGFILE3
    echo "Copy          : $copying " >> $LOGFILE3
    echo "Read parallel : $scanning" >> $LOGFILE3
    echo "Read serial   : $scanning_b" >> $LOGFILE3
    echo "Stats         : $stating " >> $LOGFILE3
    echo "Delete        : $deleting" >> $LOGFILE3    
    echo >> $LOGFILE3

    
    del_tmpfiles;
    
    echo "******* The end of mongo_multi_process *******" >> $LOGFILE

}

function set_params() {

    bytes=$1
    size=$2
    #level=$3
    #inverse=$4
    
    bytes_to_consume=$1 
    median_file_size=$2
    #files=10
    #files_total=$[ $files * $reiser_fract_tree_rep_counter ]
    #dirs_total=1
    #i=0;
    #while [ $i -lt $level ]
    #do
    #	dirs_total=$[ $dirs_total * 10 ]
    #    i=$[ $i + 1 ]
    #done
    #dirs_total=$[ $dirs_total * $reiser_fract_tree_rep_counter]
}


function run_mongo_single() {

    echo "     ***** Starting mongo_single_process, rep.N=$rep_n *****"
    echo "     ***** Starting mongo_single_process, rep.N=$rep_n *****" >> $LOGFILE

    k=0
    while [ $k -lt ${#numb_of_bytes[*]} ]
    do
	set_params ${numb_of_bytes[$k]} ${size_of_files[$k]} ${level_of_tree[$k]} ${fract_of_tree[$k]};
	par_set_n=$[ $k + 1 ]
        mongo_single_process;
	k=$[ $k + 1 ]
    done

    echo "     ***** Finished mongo_single_process, repetitions left:$rep_left *****"
    echo "     ***** Finished mongo_single_process, repetitions left:$rep_left *****" >> $LOGFILE
}

function run_mongo_multi() {

    echo "     ***** Starting mongo_multi_process, rep.N=$rep_n *****"
    echo "     ***** Starting mongo_multi_process, rep.N=$rep_n *****" >> $LOGFILE

    k=0
    while [ $k -lt ${#numb_of_bytes[*]} ]
    do
        set_params ${numb_of_bytes[$k]} ${size_of_files[$k]} ${level_of_tree[$k]} ${fract_of_tree[$k]};
	par_set_n=$[ $k + 1 ]
	mongo_multi_process;
	k=$[ $k + 1 ]
    done

    echo "     ***** Finished mongo_multi_process, repetitions left:$rep_left *****"
    echo "     ***** Finished mongo_multi_process, repetitions left:$rep_left *****" >> $LOGFILE

}

#...------------------------START MONGO BENCHMARK---------------------------

echo "***** Starting mongo.sh benchmark ******"
echo "***** Starting mongo.sh benchmark ******" >> $LOGFILE

rep_n=0
while [ $rep_left -gt 0 ]
do
    rep_n=$[ $rep_n + 1 ]
    rep_left=$[ $rep_left - 1 ]

    if [ $execute_single_process_phase -eq 1 ]
    then
        run_mongo_single;
    fi

    if [ $execute_multi_process_phase -eq 1 ]
    then
        run_mongo_multi;
    fi
done

umount $TESTDIR

echo "***** Finished a mongo.sh benchmark, repetitions:$REPETITIONS  *****"
echo "***** Finished a mongo.sh benchmark, repetitions:$REPETITIONS  *****" >> $LOGFILE
echo "Results in file : $LOGFILE"
sync
